# -*- coding: utf-8 -*-
"""Unit tests of rawhdr."""

import unittest

import numpy as np

import rawhdr


def _generate_image():
    np.random.seed(42)
    return np.random.uniform(size=(15, 10, 3))


class TestComputeScaling(unittest.TestCase):
    """Tests for rawhdr.compute_scaling()."""

    def test_unscaled(self):
        """Same brightness for both input images."""
        base_image = _generate_image()
        image = np.copy(base_image)

        result = rawhdr.compute_scaling(image, base_image)
        self.assertTrue(np.isclose(result, 1))

    def test_same_image(self):
        """id(image) == id(base_image)."""
        image = _generate_image()

        result = rawhdr.compute_scaling(image, image)
        self.assertTrue(np.isclose(result, 1))

    def test_scaled(self):
        """Basic functinality test."""
        base_image = _generate_image()

        image = base_image / 1.5
        result = rawhdr.compute_scaling(image, base_image)
        self.assertTrue(np.isclose(result, 1.5))

        image = base_image / 0.5
        result = rawhdr.compute_scaling(image, base_image)
        self.assertTrue(np.isclose(result, 0.5))

    def test_argument_check(self):
        """Test checking of function argument values."""
        base_image = _generate_image()
        image = base_image / 1.5

        rawhdr.compute_scaling(image,
                               base_image,
                               mask_width=0.3,
                               target_gamma=1.0)

        with self.assertRaises(ValueError):
            rawhdr.compute_scaling(image, base_image, mask_width=-0.01)
        with self.assertRaises(ValueError):
            rawhdr.compute_scaling(image, base_image, mask_width=1.01)
        with self.assertRaises(ValueError):
            rawhdr.compute_scaling(image, base_image, target_gamma=-1)

    def test_large_factor_fails(self):
        """Test raising of RuntimeError for very large scaling factors."""
        base_image = _generate_image()

        with self.assertRaises(RuntimeError):
            rawhdr.compute_scaling(base_image / 1000, base_image)
        with self.assertRaises(RuntimeError):
            rawhdr.compute_scaling(base_image / 1e-3, base_image)


class TestComputeWeight(unittest.TestCase):
    """Tests for rawhdr.compute_weight()."""

    def test_no_blending(self):
        """Test disabled blending."""
        image = _generate_image()

        result = rawhdr.compute_weight(image,
                                       blend_low=False,
                                       blend_high=False)
        self.assertTrue(np.all(np.isclose(result, 1)))

    def test_blending(self):
        """Test full blending."""
        linear_image = np.linspace(0, 1, 100)
        expected = np.clip((linear_image - 0.1) / 0.3, 0, 1)

        result = rawhdr.compute_weight(linear_image**2.0,
                                       blend_low=True,
                                       blend_high=False,
                                       blend_width=0.3,
                                       blend_cap=0.1,
                                       target_gamma=2.0)

        self.assertTrue(np.all(np.isclose(result, expected)))
        result = rawhdr.compute_weight(linear_image**2.0,
                                       blend_low=False,
                                       blend_high=True,
                                       blend_width=0.3,
                                       blend_cap=0.1,
                                       target_gamma=2.0)
        self.assertTrue(np.all(np.isclose(result, expected[::-1])))
        result = rawhdr.compute_weight(linear_image**2.0,
                                       blend_low=True,
                                       blend_high=True,
                                       blend_width=0.3,
                                       blend_cap=0.1,
                                       target_gamma=2.0)
        self.assertTrue(
            np.all(np.isclose(result, np.minimum(expected, expected[::-1]))))

    def test_argument_check(self):
        """Test checking of function argument values."""
        image = _generate_image()

        with self.assertRaises(ValueError):
            rawhdr.compute_weight(image, blend_width=-0.01)
        with self.assertRaises(ValueError):
            rawhdr.compute_weight(image, blend_width=0.51)
        with self.assertRaises(ValueError):
            rawhdr.compute_weight(image, blend_cap=-0.01)
        with self.assertRaises(ValueError):
            rawhdr.compute_weight(image, blend_cap=0.51)
        with self.assertRaises(ValueError):
            rawhdr.compute_weight(image, blend_width=0.3, blend_cap=0.3)
        with self.assertRaises(ValueError):
            rawhdr.compute_weight(image, target_gamma=-0.01)


class TestMergeExposures(unittest.TestCase):
    """Tests for rawhdr.merge_exposures()."""

    def test_merging(self):
        """Test merging of LDR images."""
        # Generate HDR image
        image = 2 * _generate_image()

        # Generate clipped LDR images
        scalings = [1, 0.4, 0.7, 1 / 0.7, 1 / 0.4]
        images = [np.clip(image * scaling, 1e-3, 0.95) for scaling in scalings]

        result = rawhdr.merge_exposures(images)
        self.assertTrue(np.all(np.isclose(image, result)))

    def test_argument_check(self):
        """Test checking of function argument values."""
        with self.assertRaises(ValueError):
            rawhdr.merge_exposures([])



# -*- coding: utf-8 -*-
"""RawHdr merges multiple RAW images into an floating point HDR image."""

import numpy as np

__version__ = '0.2.4'


def compute_scaling(image, base_image, mask_width=None, target_gamma=None):
    """Exposure scaling computation.

    Computes the scaling required for `image` such that its
    brightness/exposure is equal to `base_image`.

    Only the central part of the brightness distribution of both images is
    used to perform the calculation. Very bright (possibly saturated) and
    dark (possibly noisy) pixels are ignored.

    Masking of the center part can be controlled by `mask_width` and
    `target_gamma`, in most cases the default parameters should work
    reasonaly well.

    Parameters
    ----------
    image : array_like
        Image for which the scaling factor is computed.
    base_image : array_like
        Image with reference brightness/exposure.
    mask_width : float
        Size of the mask in gamma-corrected color space.
    target_gamma : float
        Target gamma used for the masking computations.

    Returns
    -------
    scaling : float
        Scaling factor such that `scaling` * `image` is approximately
        equal to `base_image`.
    """
    if mask_width is None:
        mask_width = 0.8
    if target_gamma is None:
        target_gamma = 2.2

    if mask_width <= 0 or mask_width > 1:
        raise ValueError('Maske width must be positive and at most 1.0')
    if target_gamma <= 0:
        raise ValueError('Invalid value for `target_gamma`. Must be positive')

    if base_image is image:
        return 1.0

    base_image = np.asarray(base_image)
    image = np.asarray(image)

    # Compute mask where all image data is in reasonable sensor range
    mask_min = 0.5 - 0.5 * mask_width
    mask_max = 0.5 + 0.5 * mask_width
    image_gammac = image**(1.0 / target_gamma)
    base_image_gammac = base_image**(1.0 / target_gamma)
    mask = ((base_image_gammac >= mask_min)
            & (base_image_gammac <= mask_max)
            & (image_gammac >= mask_min)
            & (image_gammac <= mask_max))

    if np.sum(mask) == 0:
        raise RuntimeError('No values to match exposure, '
                           'try increasing `mask_width`')

    # Compute scaling of `image` image to match exposure of `base_image`
    return np.mean(base_image[mask] / image[mask])


def compute_weight(image,
                   blend_low=True,
                   blend_high=True,
                   blend_width=None,
                   blend_cap=None,
                   target_gamma=None):
    """Compute per-pixel blending weights.

    Parameters
    ----------
    image : array_like
        Image for which the blending weights are computed.
    blend_low : bool
        Enables blending of the lower (dark) part of the image brightness.
    blend_high : bool
        Enabled blending of the higher (dark) part of the image brightness.
    blend_width : float
        Width of the blended regions at both ends of the brightness range.
    blend_cap : float
        Cap of dark and bright regions.
    target_gamma : float
        Gamma at which the blending is performed.

    Returns
    -------
    mask : numpy.ndarray
        Mask with the same shape as `image`.
    """
    if blend_width is None:
        blend_width = 0.2
    if blend_cap is None:
        blend_cap = 0.1
    if target_gamma is None:
        target_gamma = 2.2

    if not 0 <= blend_width <= 0.5:
        raise ValueError(
            'Invalid value for `blend_width`. Must be in range [0, 0.5]')
    if not 0 <= blend_cap <= 0.5:
        raise ValueError(
            'Invalid value for `blend_cap`. Must be in range [0, 0.5]')
    if blend_width + blend_cap > 0.5:
        raise ValueError('Invalid value for `blend_width` + `blend_cap`.'
                         'Sum must be less than 0.5')
    if target_gamma <= 0:
        raise ValueError('Invalid value for `target_gamma`. Must be positive')

    # gamma-corrected image
    image_gammac = image**(1.0 / target_gamma)
    mask = np.ones_like(image)
    if blend_low:
        # blend out dark pixels
        mask = np.minimum(np.clip((image_gammac - blend_cap) / blend_width, 0,
                                  1),
                          mask,
                          out=mask)
    if blend_high:
        # blend out bright pixels
        mask = np.minimum(np.clip((1 - image_gammac - blend_cap) / blend_width,
                                  0, 1),
                          mask,
                          out=mask)

    return mask


def merge_exposures(exposures,
                    mask_width=None,
                    blend_width=None,
                    blend_cap=None,
                    target_gamma=None):
    """Merge multiple LDR images into a HDR image.

    Parameters
    ----------
    exposures : iterable of array_like
        Floating point input images in linear color space.
    mask_width : float
        Size of the mask in gamma-corrected color space.
    blend_width : float
        Width of the blended regions at both ends of the brightness range.
    blend_cap : float
        Cap of dark and bright regions.
    target_gamma : float
        Target gamma used for the masking computations.

    Returns
    -------
    merged : numpy.ndarray
        Merged HDR image width same exposure as the first image in the
        `exposures` input list.
    """
    exposures = list(exposures)
    if not exposures:
        raise ValueError('At least one input image is required')

    # blend scalings of all images to match first image exposure
    scalings = [
        compute_scaling(exposure,
                        base_image=exposures[0],
                        mask_width=mask_width,
                        target_gamma=target_gamma) for exposure in exposures
    ]
    min_scaling, max_scaling = min(scalings), max(scalings)

    # compute blending weights
    weights = [
        compute_weight(exposure,
                       blend_low=(scaling != min_scaling),
                       blend_high=(scaling != max_scaling),
                       blend_width=blend_width,
                       blend_cap=blend_cap,
                       target_gamma=target_gamma)
        for exposure, scaling in zip(exposures, scalings)
    ]

    # normalize blending weights
    total_weight = sum(weights) + 1e-8
    for weight in weights:
        weight /= total_weight

    # blend scaled images by weight
    return sum(
        exposure * scaling * weight
        for exposure, scaling, weight in zip(exposures, scalings, weights))



# -*- coding: utf-8 -*-
"""Setup file for rawhdr."""

import os
import re
import setuptools


def read_file(*path):
    """Read file content."""
    package_path = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(package_path, *path), 'r') as text_file:
        return text_file.read()


def long_description():
    """Read long description from README.md."""
    return read_file('README.md')


def version():
    """Parse version info."""
    initfile_content = read_file('rawhdr', '__init__.py')
    match = re.search(r"__version__ = [']([^']*)[']", initfile_content, re.M)
    if match:
        return match.group(1)
    raise RuntimeError('Unable to find version string')


setuptools.setup(
    name='rawhdr',
    version=version(),
    author='Felix Thaler',
    author_email='felix.thaler@nummi.ch',
    description='A simple HDR image merger',
    long_description=long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/fthaler/rawhdr',
    packages=setuptools.find_packages(),
    scripts=['bin/rawhdrmerge'],
    install_requires=['click', 'imageio', 'numpy', 'rawpy'],
    classifiers=[
        'Programming Language :: Python', 'License :: OSI Approved :: '
        'GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: OS Independent',
        'Topic :: Multimedia :: Graphics :: Graphics Conversion'
    ],
)
